#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_metropack
----------------------------------

Tests for `metropack` module.
"""

import unittest
from metropack import pack


class TestPack(unittest.TestCase):

    def test_pack(self):
        pass
